# -*- coding: utf-8 -*-
"""
Created on 2022/09/18 16:19:57

@File -> cond_entropy.py

@Author: luolei

@Email: dreisteine262@163.com

@Describe: 条件熵
"""

import numpy as np

from .marg_entropy import MargEntropy
from ..mutual_info.mi import MutualInfoGIEF


class CondEntropy(object):
    """条件熵"""
    
    def __init__(self, x: np.ndarray, xtype: str, z: np.ndarray, ztype: str):
        # note 此处不再对x, z进行标准化和类型检查, 因为这些步骤会在底层MargEntropy和MutualInfo中进行
        self.x = x.copy().reshape(len(x), -1)
        self.z = z.copy().reshape(len(z), -1)
        self.xtype = xtype
        self.ztype = ztype
        
    def __call__(self, k=3, metric="chebyshev"):
        return MargEntropy(self.x, self.xtype)(k=k, metric=metric) -\
            MutualInfoGIEF(self.x, self.xtype, self.z, self.ztype)(k=k, metric=metric)


# -*- coding: utf-8 -*-
"""
Created on 2023/06/05 15:44:42

@File -> td_assoc_detection.py

@Author: luolei

@Email: dreisteine262@163.com

@Describe: 时延关联检测
"""

import matplotlib.pyplot as plt
import numpy as np
import logging

from ..util import build_td_series
from .surrog_indep_test import exec_surrog_indep_test


def detect_time_delayed_assoc(x, y, taus, xtype="c", ytype="c", max_size_bt=1000, method="MI-GIEF", 
                              show=False, **kwargs):
    """双变量间的时延关联检测

    :param x: x样本, np.ndarray
    :param y: y样本, np.ndarray
    :param taus: 待计算时延数组, np.ndarray
    :param xtype: x的数值类型, str, defaults to "c"
    :param ytype: y的数值类型, str, defaults to "c"
    :param max_size_bt: 重采样计算中的最大样本数, defaults to 1000
    :param method: 所使用的关联度量方法, defaults to "MI-GIEF"
    :param show: 是否显示结果, bool, defaults to False
    :param **kwargs: 如alpha, rounds等, 见exec_surrog_indep_test中的参数设置
    """
    x, y = x.flatten(), y.flatten()
    N = len(x)

    # 检查样本长度
    try:
        assert N == len(y)
    except Exception as e:
        raise RuntimeError("x的长度与y不相等") from e

    # 抽样样本量参数设置, 后续计算需以此为基础统一计算
    N_drop_tau_max = N - np.max(np.abs(taus))  # 时延计算中的最低样本量
    if N_drop_tau_max <= max_size_bt:
        size_bt = N_drop_tau_max
    else:
        logging.warning(f"采用默认max_size_bt={max_size_bt}作为size_bt, 用于限制代用数据重采样规模")
        size_bt = max_size_bt
    
    # 逐时延计算
    td_assocs = np.array([])
    td_indeps = np.array([])
    for tau in taus:
        # 构造时延序列
        x_td, y_td = build_td_series(x, y, tau)
        
        # 进行关联计算
        assoc, (_, indep, _) = exec_surrog_indep_test(
            x_td, y_td, method, xtype=xtype, ytype=ytype, size_bt=size_bt, **kwargs)
        
        td_assocs = np.append(td_assocs, assoc)
        td_indeps = np.append(td_indeps, indep)
        
    # 显示结果, 红色点表示显著
    if show:
        _show_results(taus, td_assocs, td_indeps)
        
    return td_assocs, td_indeps
        

def _show_results(taus, td_assocs, td_indeps):
    plt.figure()
    plt.scatter(
        taus, td_assocs, edgecolors=["k" if p==1 else "r" for p in td_indeps], c="w", lw=2, zorder=1)
    plt.plot(taus, td_assocs, c="k", lw=1.5, zorder=0)
    plt.grid(alpha=0.3, zorder=-1)
    plt.xlabel("$\\tau$")
    plt.ylabel("MI")


from .entropy.marg_entropy import MargEntropy
from .entropy.cond_entropy import CondEntropy
from .mutual_info.mi import MutualInfoGIEF
from .mutual_info.cmi import CondMutualInfoGIEF

__all__ = ["MargEntropy", "CondEntropy", "MutualInfoGIEF", "CondMutualInfoGIEF"]


import numpy as np

from ..setting import DTYPES

from .gief import MutualInfoGIEF, CondMutualInfoGIEF, MargEntropy, CondEntropy
from .kde import MutualInfoKDE
from .mic import MIC, CMIC
from .model_based import MutualInfoModel, CondMutualInfoModel
from .quant_based import MutualInfoClassic, MutualInfoDarbellay
from .correlation_coeff import cal_dist_corr, cal_pearson_corr, cal_spearman_corr

ASSOC_METHODS = [
    "PearsonCorr", "SpearmanCorr", "DistCorr",
    "MI-GIEF", "MI-model", "MI-cut", "MI-qcut", "MI-Darbellay", "MI-KDE",
    "MIC", "RMIC"
]
COND_ASSOC_METHODS = [
    "CMI-GIEF", "CMI-model", "CMI-cut", "CMI-qcut",
    "CMIC", "CRMIC"
]


def cal_marg_entropy(x: np.ndarray, xtype: str, **kwargs):
    """计算边际熵

    :param x: x数据, 一维或多维
    :param xtype: x的数值类型, "d"或者"c"
    :param kwargs:
        k: int, 当xtype == "c"时选用的KNN近邻数, defaults to 3
        metric: str, 当xtype == "c"时选用的距离度量方式, defaults to "chebyshev"
    """
    assert xtype in DTYPES
    return MargEntropy(x, xtype)(**kwargs)


def cal_cond_entropy(x: np.ndarray, xtype: str, z: np.ndarray, ztype: str, **kwargs):
    """计算条件熵

    :param x: x数据, 一维或多维
    :param xtype: x的数值类型, "d"或者"c"
    :param z: z数据, 一维或多维
    :param ztype: z的数值类型, "d"或者"c"
    :param kwargs:
        k: int, 当xtype == "c"时选用的KNN近邻数, defaults to 3
        metric: str, 当xtype == "c"时选用的距离度量方式, defaults to "chebyshev"
    """
    assert xtype in DTYPES
    assert ztype in DTYPES
    return CondEntropy(x, xtype, z, ztype)(**kwargs)


def cal_assoc(x, y, method, xtype=None, ytype=None, **kwargs):
    """计算关联系数"""
    assert method in ASSOC_METHODS
    
    # 线性相关系数
    if method == "PearsonCorr":
        return cal_pearson_corr(x, y)
    elif method == "SpearmanCorr":
        return cal_spearman_corr(x, y)
    elif method == "DistCorr":
        return cal_dist_corr(x, y)
    
    # 互信息
    elif method == "MI-GIEF":
        return MutualInfoGIEF(x, xtype, y, ytype)(**kwargs)  # kwargs: k, metric
    elif method == "MI-model":
        return MutualInfoModel(x, xtype, y, ytype)(**kwargs)  # kwargs: model, test_ratio
    elif method == "MI-cut":
        return MutualInfoClassic(x, y)(method="cut")
    elif method == "MI-qcut":
        return MutualInfoClassic(x, y)(method="qcut")
    elif method == "MI-Darbellay":
        return MutualInfoDarbellay(x, y)()
    elif method == "MI-KDE":
        return MutualInfoKDE(x, y)()
    
    # 最大信息系数
    elif method == "MIC":
        return MIC(x, y)(method="mic")
    elif method == "RMIC":
        return MIC(x, y)(method="rmic", **kwargs)
    else:
        raise ValueError(f"unsupported method {method}")
    
    
def cal_cond_assoc(x, y, z, method, xtype=None, ytype=None, ztype=None, **kwargs):
    """计算条件关联系数"""
    assert method in COND_ASSOC_METHODS
    
    # 条件互信息
    if method == "CMI-GIEF":
        return CondMutualInfoGIEF(x, xtype, y, ytype, z, ztype)(**kwargs)  # kwargs: k, metric, method for estimating CMI
    elif method == "CMI-model":
        return CondMutualInfoModel(x, xtype, y, ytype, z, ztype)(**kwargs)  # kwargs: model, test_ratio
    elif method == "CMI-cut":
        return MutualInfoClassic(x, y, z)(method="cut")
    elif method == "CMI-qcut":
        return MutualInfoClassic(x, y, z)(method="qcut")
    
    # 条件最大信息系数
    elif method == "CMIC":
        return CMIC(x, y, z)(method="mic")
    elif method == "CRMIC":
        return CMIC(x, y, z)(method="rmic")
    else:
        raise ValueError(f"unsupported method {method}")